/***
 * Juliac
 * Copyright (C) 2011-2013 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.opt.ultracomp;

import java.lang.reflect.Modifier;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.control.lifecycle.InvocationCounterItf;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.objectweb.fractal.juliac.ucf.UnifiedMethod;
import org.objectweb.fractal.juliac.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.ClassSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.ThenSourceCodeVisitor;

/**
 * <p>
 * This class generates the source code of lifecycle interceptors.
 * </p>
 * 
 * <p>
 * This class specializes the original Juliac lifeccyle interceptor genarator by
 * always invoking the <code>incrementFcInvocationCounter()</code> and
 * <code>decrementFcInvocationCounter()</code> methods instead of accessing
 * directly the <code>fcInvocationCounter</code> field. This is needed, not only
 * when interceptors are componentized, but also for object interceptors that
 * access controllers implemented in Scala (in this last case, it appears that
 * public fields are privatized when compiled into bytecode by Scala and are
 * then no longer accessible from the generated Java object interceptor.)
 * </p>
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.5
 */
public class LifeCycleSourceCodeGenerator
extends org.objectweb.fractal.juliac.proxy.LifeCycleSourceCodeGenerator {
        
    public LifeCycleSourceCodeGenerator() {
    	super();
    }
    
    public LifeCycleSourceCodeGenerator(
        Juliac jc,
        InterfaceType it, MembraneDesc<?> membraneDesc, boolean mergeable ) {
        
        super(jc,it,membraneDesc,mergeable);
    }
    
   	@Override
    public String[] getImplementedInterfaceNames() {
   		String[] old = super.getImplementedInterfaceNames();
   		String[] impls = new String[old.length+1];
   		System.arraycopy(old,0,impls,0,old.length);
   		impls[old.length] = BindingController.class.getName();
    	return impls;
    }

   	@Override
    public void generateFields( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {
        cv.visitField(
    		Modifier.PRIVATE, InvocationCounterItf.class.getName(),
    		"_lc", null );
    }
        
    @Override
    public void generateMethodInitFcController( BlockSourceCodeVisitor mv, UnifiedClass proxycl ) {
    	/*
    	 * Indeed nothing.
    	 * 
    	 * Replace the overriden method body. The idea is that _lc is now set
    	 * via a bind by the membrane initializer.
    	 */
    }

    @Override
    public void generateMethodOthers( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {
    	
    	super.generateMethodOthers(cv,proxycl);
    	
    	// TODO This concern should be implemented in another generator
    	
    	final String iccname = "/invocation-counter-controller";
    	final String implname =
			FCUltraCompCtrlSourceCodeGenerator.SYM_ITF_PREFIX +
    		it.getFcItfName();
    	
    	BlockSourceCodeVisitor mv =
			cv.visitMethod(
				Modifier.PUBLIC, null, "void", "bindFc",
				new String[]{ "String clientItfName", "Object serverItf" },
				null );
    	ThenSourceCodeVisitor thenv = mv.visitIf("clientItfName.equals(\""+iccname+"\")");
    	thenv.visitSet("_lc","("+InvocationCounterItf.class.getName()+") serverItf");
    	thenv.visitEnd();
    	thenv = mv.visitIf("clientItfName.equals(\""+implname+"\")");
    	thenv.visitSet("_impl","("+proxycl.getName()+") serverItf");
    	thenv.visitEnd();
    	mv.visitEnd();

    	mv = cv.visitMethod(
				Modifier.PUBLIC, null, "String[]", "listFc", null, null );
    	mv.visitIns("return new String[]{\""+iccname+"\",\""+implname+"\"}");
    	mv.visitEnd();
    	
    	mv = cv.visitMethod(
				Modifier.PUBLIC, null, "Object", "lookupFc",
				new String[]{ "String clientItfName" },
				new String[]{NoSuchInterfaceException.class.getName()} );
    	thenv = mv.visitIf("clientItfName.equals(\""+iccname+"\")");
    	thenv.visitIns("return _lc");
    	thenv.visitEnd();
    	thenv = mv.visitIf("clientItfName.equals(\""+implname+"\")");
    	thenv.visitIns("return _impl");
    	thenv.visitEnd();
    	mv.visitIns(
			"throw new ",
			NoSuchInterfaceException.class.getName(),"(clientItfName)");
    	mv.visitEnd();
    	
    	mv = cv.visitMethod(
				Modifier.PUBLIC, null, "void", "unbindFc",
				new String[]{ "String clientItfName" },
				null );
    	thenv = mv.visitIf("clientItfName.equals(\""+iccname+"\")");
    	thenv.visitSet("_lc","null");
    	thenv.visitEnd();
    	thenv = mv.visitIf("clientItfName.equals(\""+implname+"\")");
    	thenv.visitSet("_impl","null");
    	thenv.visitEnd();
    	mv.visitEnd();
    }
        
    @Override
    public void generateProxyMethodBodyBeforeCode(
        BlockSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym ) {
        
    	BlockSourceCodeVisitor sync = mv.visitSync("_lc");
    	sync.visitIns("_lc.incrementFcInvocationCounter()");
    	sync.visitEnd();
        mv.visitln("      try {");
    }
    
    @Override
    public void generateProxyMethodBodyFinallyCode(
        BlockSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym ) {
        
        mv.visitln("      }");
        mv.visitln("      finally {");
    	BlockSourceCodeVisitor sync = mv.visitSync("_lc");
    	sync.visitIns("_lc.decrementFcInvocationCounter()");
    	sync.visitEnd();
        mv.visitln("      }");
    }
}
